@@ -8,6 +8,7 @@
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE CPP #-}
 
 -- | API for database actions. The API deals with fields and entities.
 -- In SQL, a field corresponds to a column, and should be a single non-composite value.
@@ -68,7 +69,7 @@ import qualified Data.Enumerator.List as EL
 import qualified Control.Monad.IO.Class as Trans
 
 import qualified Control.Exception as E
-import Control.Monad.Error.Class (Error (..))
+import Control.Monad.Trans.Error (Error (..))
 
 import Data.Bits (bitSize)
 import Control.Monad (liftM)
@@ -77,9 +78,15 @@ import qualified Data.Text.Encoding as T
 import qualified Data.Text.Encoding.Error as T
 import Web.PathPieces (SinglePiece (..))
 import qualified Data.Text.Read
-import Data.Maybe (fromMaybe, isJust)
+import Control.Monad.IO.Class (MonadIO)
 
+#if MIN_VERSION_monad_control(0, 3, 0)
+import Control.Monad.Trans.Control (MonadBaseControl)
+#define MBCIO MonadBaseControl IO
+#else
 import Control.Monad.IO.Control (MonadControlIO)
+#define MBCIO MonadControlIO
+#endif
 import Data.Object (TextObject)
 
 fst3 :: forall t t1 t2. (t, t1, t2) -> t
@@ -500,15 +507,12 @@ checkUnique val =
             Just _ -> return False
 
 limitOffsetOrder :: PersistEntity val => [SelectOpt val] -> (Int, Int, [SelectOpt val])
-limitOffsetOrder opts = let (l,o,ord) = go opts (Nothing, Nothing, []) in
-    (fromMaybe 0 l, fromMaybe 0 o, ord)
+limitOffsetOrder opts =
+    foldr go (0, 0, []) opts
   where
-    go []              tup = tup
-    go (LimitTo  x:xs) (_,o,ord) = let tup = (Just x, o, ord) in
-      if isJust o then tup else go xs tup
-    go (OffsetBy x:xs) (l,_,ord) = let tup = (l, Just x, ord) in
-      if isJust l then tup else go xs tup
-    go (x:xs)          (l, o, ord) = go xs (l,o,x:ord)
+    go (LimitTo l) (_, b, c) = (l, b ,c)
+    go (OffsetBy o) (a, _, c) = (a, o, c)
+    go x (a, b, c) = (a, b, x : c)
 
 -- | Call 'select' but return the result as a list.
 selectList :: (PersistEntity val, PersistBackend b m)
@@ -583,7 +587,7 @@ class PersistConfig c where
     type PersistConfigPool c
     loadConfig :: TextObject -> Either String c
     -- | I really don't want Applicative here, but it's necessary for Mongo.
-    withPool :: (Applicative m, MonadControlIO m) => c -> (PersistConfigPool c -> m a) -> m a
-    runPool :: MonadControlIO m => c -> PersistConfigBackend c m a
+    withPool :: (Applicative m, MBCIO m, MonadIO m) => c -> (PersistConfigPool c -> m a) -> m a
+    runPool :: (MBCIO m, MonadIO m) => c -> PersistConfigBackend c m a
             -> PersistConfigPool c
             -> m a
